"""Spin up a basic server endpoint."""

import asyncio
import vizible
from . import MutableHostedLLM

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#                                           Constants.                                            #
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
_HOST = "127.0.0.1"
_PORT = 8000
_ADDR = f"http://{_HOST}:{_PORT}/static/"

async def main():
    """Main program loop, run asychnronously."""
    server = MutableHostedLLM(host=_HOST, port=_PORT)
    try:
        await server.up()
        vizible.green("Successfully launched server and enter eval framework")
        vizible.blue(f"Server address: \n{_ADDR}")
        await asyncio.sleep(24 * 60 * 60)  # 1 day.
    except KeyboardInterrupt:
        vizible.red("Turn down requested.")
    finally:
        await server.down()
    vizible.green("Async server terminated successfully if you can read this!**")
    vizible.green("\n\n\t\t**(assumes that you have the ability to read)")


if __name__ == "__main__":
    asyncio.run(main())



"""Stateful LLM."""

from typing import AsyncIterable, List, Tuple

import functools
import threading
import tqdm

import immutabledict
from mlx import optimizers
from mlx import nn
import mlx.core as mx
from mlx_lm.generate import generate, stream_generate
from mlx_lm.utils import load
from mlx_lm.tuner.utils import linear_to_lora_layers
from transformers.tokenization_utils import PreTrainedTokenizer
import vizible

from .libs import revise
from ._classes import InteractionHistory, TrainingExample


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#                         Default constants.                          #
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
_MODEL_NAME = "lmstudio-community/DeepSeek-R1-0528-Qwen3-8B-MLX-4bit"
# _MODEL_NAME = "mlx-community/DeepSeek-R1-Distill-Qwen-1.5B-4bit"
_MAX_TOKENS = 32768
_LEARNING_RATE = 5e-5
_EPOCHS = 5
_NUM_LORA_LAYERS = 16
_LORA_PARAMETERS = immutabledict.immutabledict(
    {"rank": 8, "dropout": 0.0, "scale": 20.0}
)
_USE_DORA = False


def _load(
    model_name: str,
    num_lora_layers: int,
    use_dora: bool,
    lora_parameters: dict | None = None,
) -> Tuple[nn.Module, PreTrainedTokenizer]:
    """Load model parameters and tokenizer.

    Args:
        model_name: Path or Huggingface name.
        num_lora_layers: Number of LORA layers, if LORA is enabled.
        use_dora: Whether to use DORA, if LORA is enabled.
        lora_parameters: LORA hyperparameters. If not None, LORA will be enabled.

    Returns:
        Model and tokenizer.
    """
    model, wrapped_tokenizer = load(model_name)
    print("Freezing all non-Lora model parameters.")
    model.freeze()
    if lora_parameters is not None:
        linear_to_lora_layers(
            model=model,
            num_layers=num_lora_layers,
            config=lora_parameters,
            use_dora=use_dora,
        )
    return model, wrapped_tokenizer._tokenizer  # pylint: disable=protected-access


def _loss_fn(
    model: nn.Module,
    inputs: mx.array,
    targets: mx.array,
    mask: mx.array,
) -> mx.array:
    logits: mx.array = model(inputs, mx.ones_like(inputs))
    loss = nn.losses.cross_entropy(logits, targets, reduction="mean") * mask
    normalized_loss = loss.sum() / mask.sum()
    return normalized_loss


class StatefulLLM:
    """Model container that bundles revision, learning, and serving logic."""

    def __init__(
        self,
        model_name: str = _MODEL_NAME,
        learning_rate: float = _LEARNING_RATE,
        max_tokens: int = _MAX_TOKENS,
        epochs: int = _EPOCHS,
        num_lora_layers: int = _NUM_LORA_LAYERS,
        lora_parameters: dict | None = _LORA_PARAMETERS,
        use_dora: bool = _USE_DORA,
    ) -> None:
        """Initializes the StatefulLLM

        Args:
            model_name: Path or Huggingface name.
            learning_rate: Backpropagation hyperparameter.
            max_tokens: Maximum number of tokens to decode in a single turn.
            epochs: Number of training epochs to perform on self-reflective model revisions.
            num_lora_layers: Number of LORA layers, if LORA is enabled.
            lora_parameters: LORA hyperparameters. If not None, LORA will be enabled.
            use_dora: Whether to use DORA, if LORA is enabled.

        Returns: None
        """
        self._lock = threading.Lock()
        with self._lock:
            self._messages: list[dict[str, str]] = []
            self._model, self._tokenizer = _load(
                model_name,
                num_lora_layers=num_lora_layers,
                lora_parameters=lora_parameters,
                use_dora=use_dora,
            )
        self._model_name = model_name
        self._optimizer = optimizers.AdamW(learning_rate=learning_rate)
        self._epochs = epochs
        self._max_tokens = max_tokens
        self._bos_token = self._tokenizer.special_tokens_map.get("bos_token", "")
        self._eos_token = self._tokenizer.special_tokens_map.get("eos_token", "")

        self._model_is_stable = True
        self._response_stream = None

    @property
    def ok(self) -> bool:
        """Checks if the model is in a stable state (i.e. not in the middle of backprop)."""
        return self._model_is_stable

    def generate_response(
        self,
        prompt: str,
        use_history: bool = True,
        max_tokens: int | None = None,
    ) -> str:
        """Generates model response, including handling tokenization and de-tokenization.

        Args:
            prompt: User input.
            use_history: Whether to include previous interaction history within the prompt.
            max_tokens: Maximum number of tokens to decode in a single turn.

        Returns:
            Model-generated response.
        """
        vizible.blue(f"Generating response for {prompt = }")
        message = {"role": "user", "content": prompt}
        if max_tokens is None:
            max_tokens = self._max_tokens
        if use_history:
            self._messages.append(message)
            messages = self._messages
        else:
            messages = [message]
        tokenized_prompt = self._tokenizer.apply_chat_template(messages)
        model_response = generate(
            self._model,
            self._tokenizer,
            prompt=tokenized_prompt,
            verbose=True,
            max_tokens=max_tokens,
        )
        return model_response.strip()

    async def stream_response(
        self,
        prompt: str,
        use_history: bool = True,
        max_tokens: int | None = None,
    ) -> AsyncIterable[str]:
        """Stream generated response"""
        message = {"role": "user", "content": prompt}
        if max_tokens is None:
            max_tokens = self._max_tokens
        print(f'{self._messages = }')
        if use_history:
            self._messages.append(message)
            messages = self._messages
        else:
            messages = [message]
        tokenized_prompt = self._tokenizer.apply_chat_template(messages)
        responses = []
        try:
            for response in stream_generate(
                self._model,
                self._tokenizer,
                prompt=tokenized_prompt,
                max_tokens=max_tokens,
            ):
                text = response.text
                responses.append(text)
                yield text
        finally:
            self._response_stream = "".join(responses)

    def _tokenize(self, inp: str, dtype: mx.Dtype = mx.int32) -> mx.array:
        return mx.array(
            self._tokenizer.encode(inp, add_special_tokens=False), dtype=dtype
        )

    def _self_correct(
        self,
        interaction_history: List[InteractionHistory],
        indices_to_review: List[int] | None,
        verbose: bool,
    ) -> TrainingExample:
        self._model_is_stable = False
        vizible.green("\n--- Starting Self-Correction and Training Cycle ---")
        if indices_to_review is None:
            indices_to_review = list(range(len(interaction_history)))
        if not interaction_history or not indices_to_review:
            raise ValueError("No unreviewed interactions to process.")
        if verbose:
            vizible.magenta(
                f"Found {len(interaction_history)} unreviewed interactions."
            )
        interactions_to_review = []
        for idx in indices_to_review:
            # Mark as reviewed to skip re-processing in the next cycle.
            interaction_history[idx].reviewed = True
            interactions_to_review.append(interaction_history[idx])

        # 1. Have the model re-evaluate its past responses and try to improve upon one of its turns.
        review_prompt = revise.make_revision_prompt(
            interactions_to_review, self._tokenizer
        )
        llm_rewrite_response = self.generate_response(review_prompt, use_history=False)
        if verbose:
            vizible.blue(f"  - Response: {llm_rewrite_response}")

        # 2. Prepare training data to train the model on how it should have responded in this
        #    situation.
        example = revise.make_collated_training_example(
            llm_rewrite_response, interactions_to_review, self._tokenizer
        )
        return example

    def _train(self, example: TrainingExample, verbose: bool):
        state = [self._model.state, self._optimizer.state, mx.random.state]
        mx.eval(state)
        loss_and_grad_fn = nn.value_and_grad(self._model, _loss_fn)

        @functools.partial(mx.compile, inputs=state, outputs=state)
        def _step(inputs, labels, mask):
            loss, grads = loss_and_grad_fn(self._model, inputs, labels, mask)
            self._optimizer.update(self._model, grads)
            return loss

        # 3. Train the model on the new, improved examples (backward-pass)
        losses = []

        if verbose:
            vizible.green(f"During training: {mx.metal.device_info() = }")
        mx.set_wired_limit(mx.metal.device_info()["max_recommended_working_set_size"])
        mx.set_cache_limit(mx.metal.device_info()["max_buffer_length"])
        world = mx.distributed.init()
        world_size = world.size()
        rank = world.rank()
        if world_size > 1:
            tqdm.tqdm.write(f"Node {rank} of {world_size}")

        self._model.train(True)
        for epoch in tqdm.tqdm(
            range(self._epochs), desc="Training", total=self._epochs
        ):
            loss = _step(example.input, example.label, example.mask)
            mx.eval(state, loss)
            if verbose:
                vizible.green(f"Epoch: {epoch}\tLoss: {loss = }")
            losses.append(loss)

        self._model.train(False)
        if verbose:
            vizible.cyan(f"{losses = }")

    def self_correct_and_train(
        self,
        interaction_history: List[InteractionHistory],
        indices_to_review: List[int] | None = None,
        verbose: bool = False,
    ) -> bool:
        """Cycle in which the model revises a previously unreviewed prompt and trains from its rewrite.

        Args:
            interaction_history: Past user and model messages.
            indices_to_review: Optional indices of relevant interactions to revise within
                               interaction_history. If not set, all interactions will be reviewed.
            verbose: Enable verbose logging.

        Returns:
            Whether the process completed successfully.
        """
        self._model_is_stable = False

        # Prepare training example from self-reflective revision of past dialog.
        example = self._self_correct(interaction_history, indices_to_review, verbose)

        # Train the model on the new, improved examples (backward-pass)
        self._train(example, verbose)

        self._model_is_stable = True
        return True



"""Common class definitions."""

import dataclasses
import mlx.core as mx
from pydantic import BaseModel


@dataclasses.dataclass
class InteractionHistory:
    """Event turn during user-LLM dialog

    Attributes:
        idx: Index of current turn amongst all global turns.
        user_input: User-provided prompt.
        llm_response: LLM response.
        reviewed: Whether this interaction has been reviewed already.
        timestamp: When the interaction took place, measured in seconds.
    """
    idx: int
    user_input: str
    llm_response: str = ""
    reviewed: bool = False
    timestamp: float = 0.0

@dataclasses.dataclass
class TrainingExample:
    """Pre-tokenized training data

    Attributes:
        user_input: Actual model response.
        label: Target model response.
        mask: Mask of model response that dictates which parts are used in training.
    """
    input: mx.array
    label: mx.array
    mask: mx.array


class InteractionRequest(BaseModel):
    """User prompt to be sent to the LLM.

    Attributes:
        prompt: User-provided input.
    """
    prompt: str


class InteractionResponse(BaseModel):
    """LLM response to user-provided prompt
    
    Attributes:
        response: LLM-generated text response.
        interaction_id: Index of the current response within the context of the current session.
    """
    response: str
    interaction_idx: int


class ReviewResponse(BaseModel):
    """Response to initiating asynchronous review of entire unreviewed interaction history.

    Attributes:
        message: Human-readable output after completion of review.
        unreviewed_count: the number of unreviewed interactions handled by this operation.
    """
    message: str
    unreviewed_count: int


class SyncResponse(BaseModel):
    """Response after server completes all unfinished background tasks.

    Attributes:
        message: Human-readable message.
        tasks_count: Number of tasks waited on and successfully finished.
        elapsed_time: Amount of time background tasks took to finish.
    """
    message: str
    tasks_count: int
    elapsed_time: float



"""Demo to showcase interacting with a stateful model in a web browser."""

import asyncio
import logging
import sys

import vizible
import adaptible


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#                                           Constants.                                            #
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
_HOST = "127.0.0.1"
_PORT = 8000
_ADDR = f"{_HOST}:{_PORT}/static/"


async def main():
    """Main program loop, run asychnronously."""
    # server = adaptible.MutableHostedLLM(host=_HOST, port=_PORT)
    server = adaptible.MutableHostedLLM(host=_HOST, port=_PORT)
    try:
        await server.up()
        vizible.green("Successfully launched server and enter eval framework")
        vizible.magenta(f"Navigate to: \nhttp://{_ADDR}")
        while True:
            await asyncio.sleep(1)
    except Exception as e:
        raise ValueError("Serving failed.") from e
    finally:
        await server.down()
        vizible.green("Async server terminated successfully if you can read this!**")
        vizible.green("\n\n\t\t**(assumes that you have the ability to read)")


if __name__ == "__main__":
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        stream=sys.stdout,
    )
    logger = logging.getLogger(__name__)  # Get a logger for this module
    asyncio.run(main())



"""Utilities for self-reflective model revisions."""

import copy
import re
from typing import Sequence, Tuple

import mlx.core as mx
from transformers.tokenization_utils import PreTrainedTokenizer

from .._classes import InteractionHistory, TrainingExample

_REWRITE_INSTRUCTIONS = (
    "You are a professional editor that coaches people and LLMs on how to improve their "
    "conversational skills."
    "Today, you have been asked to inspect a dialog between a human user and an LLM and revise "
    "one of the LLM's responses."
    "It is important to consider the full context of the conversation, especially if the LLM's "
    "responses were not satisfactory during the dialog."
    "Be sure that your revised response(s) would be a realistic response considering that the "
    "later context of the conversation was not yet known."
    "Your response MUST:"
    '    * Choose an "assistant" response to revise. Label your rewritten response with the '
    "dialog turn that it relates to by starting your response with '''[[<X>]] "
    "(e.g. '''[[0]], '''[[1]], etc.)."
    "    * End your response with [[/<X>]]'''. This separator is imperative so please do not "
    "forget it."
    "    * Since you are revising one of the existing responses, make sure your '''[[<X>]] label "
    "corresponds with a dialog index that already exists within the original dialog."
)


def _make_revision_prompt(
    past_dialog: str, instructions: str = _REWRITE_INSTRUCTIONS
) -> str:
    """Create an LLM prompt for the model to critique and rewrite a previous response.

    Args:
        past_dialog: Relevant conversation history pre-serialized to a string.
        instructions: Prompt pre-amble containing instructions on how to revise the model response.

    Returns: a formatted text prompt.
    """
    return "\n".join(
        (
            instructions,
            "<PAST_DIALOG>",
            past_dialog,
            "</PAST_DIALOG>",
        )
    )


def _isolate_turn_to_rewritten_turn_index(model_response: str) -> int:
    rewritten_indices = re.findall(r"\[\[([0-9]*)\]\]", model_response)
    unique_rewritten_indices = set(rewritten_indices)
    index_to_rewrite = min(map(int, unique_rewritten_indices))

    if not rewritten_indices:
        raise ValueError(f"Failed to parse a turn ID from {model_response}")
    return index_to_rewrite


def _parse_rewritten_response(model_response: str, idx: int) -> str:
    sor_index = None
    eor_index = None
    for group in re.finditer(rf"\[\[{idx}\]\]", model_response):
        sor_index = group.end()
    for group in re.finditer(rf"\[\[/{idx}\]\]", model_response):
        eor_index = group.start()
    return model_response[sor_index:eor_index].strip()


def _serialize_interactions_to_string(
    interactions: Sequence[InteractionHistory],
    should_enumerate: bool,
    tokenizer: PreTrainedTokenizer,
    continue_final_message: bool,
) -> Tuple[str, Sequence[str]]:
    turns = []
    for interaction in interactions:
        messages = [
            {
                "role": "user",
                "content": interaction.user_input,
            },
            {
                "role": "assistant",
                "content": interaction.llm_response,
            },
        ]
        turn = tokenizer.apply_chat_template(
            conversation=messages,
            tokenize=False,
            continue_final_message=continue_final_message,
        )
        turns.append(turn)
    if should_enumerate:
        turns_as_text = "\n".join(
            [f"[[{idx}]]{turn}" for idx, turn in enumerate(turns)]
        )
    else:
        turns_as_text = "\n".join(turns)
    return turns_as_text, turns


def _pad(arr: mx.array, max_len: int, padding_token: int) -> mx.array:
    return mx.pad(
        arr,
        (0, max_len - len(arr)),
        mode="constant",
        constant_values=padding_token,
    )


def _collate_fn(
    batch_data: Sequence[TrainingExample], padding_token: int = 0
) -> TrainingExample:
    # Pad sequences to the maximum length in the batch
    max_len = max(max(map(len, (d.input, d.label, d.mask))) for d in batch_data)
    padded_inputs = []
    padded_labels = []
    padded_masks = []
    for item in batch_data:
        padded_inputs.append(_pad(item.input, max_len, padding_token))
        padded_labels.append(_pad(item.label, max_len, padding_token))
        padded_masks.append(_pad(item.mask, max_len, padding_token))
    return TrainingExample(
        input=mx.stack(padded_inputs),
        label=mx.stack(padded_labels),
        mask=mx.stack(padded_masks),
    )


def make_revision_prompt(
    interactions: Sequence[InteractionHistory],
    tokenizer: PreTrainedTokenizer,
    instructions: str = _REWRITE_INSTRUCTIONS,
) -> str:
    """Create a prompt for model self-reflective revision based on past interactions.
    
    Args:
        interactions: Past interactions.
        tokenizer: Tokenizer. Not actually used here. TODO - fix.
        instructions: Revision prompt.

    Returns: 
        Formatted prompt text.
    """
    past_dialog, _ = _serialize_interactions_to_string(
        interactions=interactions,
        should_enumerate=True,
        tokenizer=tokenizer,
        continue_final_message=False,
    )
    return _make_revision_prompt(past_dialog, instructions)

def make_collated_training_example(
    response: str,
    interactions: Sequence[InteractionHistory],
    tokenizer: PreTrainedTokenizer,
    padding_token: int = 0,
) -> TrainingExample:
    """Convert past interactions and model revision response into a batched training example.
    
    Args:
        response: Model-generated revision response.
        interactions: Past interactions considered when generating the model response.
        tokenizer: Model-specific tokenizer.
        padding_token: Token to use for padding.

    Returns: a collated training example, ready for model ingestion.
    
    """
    _, turns = _serialize_interactions_to_string(
        interactions=interactions,
        should_enumerate=True,
        tokenizer=tokenizer,
        continue_final_message=False,
    )
    eos_tag = tokenizer.special_tokens_map.get("eos_token", "")
    if isinstance(eos_tag, list):
        eos_tag = eos_tag[0]
    bos_tag = tokenizer.special_tokens_map.get("bos_token", "")
    if isinstance(bos_tag, list):
        bos_tag = bos_tag[0]
    index_to_rewrite = _isolate_turn_to_rewritten_turn_index(response)
    rewritten_response = _parse_rewritten_response(response, index_to_rewrite)
    interaction_to_revise = copy.deepcopy(interactions[index_to_rewrite])
    interaction_to_revise.llm_response = ""
    revised_dialog_turn, _ = _serialize_interactions_to_string(
        interactions=[interaction_to_revise],
        should_enumerate=False,
        tokenizer=tokenizer,
        continue_final_message=True,
    )
    def _tokenize(text: str, dtype: mx.Dtype = mx.int32) -> mx.array:
        return mx.array(tokenizer.encode(text, add_special_tokens=False), dtype=dtype)


    dialog_pre_revision = _tokenize(
        "\n".join(list(turns[:index_to_rewrite]) + [revised_dialog_turn])
    )
    revision = _tokenize(bos_tag + rewritten_response + eos_tag)
    tokenized_rewritten_dialog = mx.concat([dialog_pre_revision, revision])
    mask = mx.concat([mx.zeros_like(dialog_pre_revision), mx.ones_like(revision)])
    training_example = TrainingExample(
        input=tokenized_rewritten_dialog[:-1],
        label=tokenized_rewritten_dialog[1:],
        mask=mask,
    )
    return _collate_fn([training_example], padding_token)
